/** 
 *   Davis Weather daemon
 *
 *   (c) Copyright 2005 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 *   Besed on: Davis Weather Tool - dwt 1.0  6-Jun-2004 
 *             Copyright (C) 2004 Con Tassios <ct@swin.edu.au>
 *
 *
 *   TODO: 
 **/
#define _GNU_SOURCE

#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <search.h>
#include <math.h>
#include <ctype.h>
#include "dwt.h"


#define U_STR "KPH|MPH|MPS|FAH|CEL|HPA|IHG|MIL|INC|CLK"
#define U_CMD "FRMT|TAGS|LIST|GETTIME|SETTIME|TEST|WRD|VER|RXCHECK"


static map t[] ={  
	{ "IN_TEMP",         9, 2, GET_temp,          'F',  
	  "Indoor temperature",  DEF_TM },
	{ "IN_HUMIDITY",    11, 1, GET_humidity,      'U',
	  "Indoor humidity",     0      },
	{ "OUT_TEMP",       12, 2, GET_temp,          'F',
	  "Outdoor temperature", DEF_TM },
	{ "OUT_HUMIDITY",   33, 1, GET_humidity,      'U',
	  "Outdoor humidity",    0      },
	{ "DEW_POINT",       0, 0, GET_dewpoint,      'F',
	  "Dewpoint",            DEF_TM },
	{ "WIND_CHILL",      0, 0, GET_windchill,     'F', 
	  "Wind Chill",          DEF_TM },
	{ "HEAT_INDEX",      0, 0, GET_heatindex,     'F', 
	  "Heat Index",          DEF_TM },
	{ "BAROMETER",       7, 2, GET_barometer,     'F',
	  "Barometer",           DEF_AP },
	{ "WIND_DIRECTION", 16, 2, GET_winddirection, 'S', 
	  "Wind Direction",      0      },
	{ "WIND_DEGREE",    16, 2, GET_winddegree,    'U',
	  "Wind Degree",         0      },
	{ "WIND_SPEED",     14, 1, GET_windspeed,     'U',
	  "Wind Speed",          DEF_WS },
	{ "RAIN_TODAY",     50, 2, GET_rain,          'F',
	  "Rain today",          DEF_RF },
	{ "RAIN_MONTH",     52, 2, GET_rain,          'F',
	  "Rain month",          DEF_RF },
	{ "RAIN_YEAR",      54, 2, GET_rain,          'F',
	  "Rain year",           DEF_RF },
	{ "SUNRISE",        91, 2, GET_time,          'U',
	  "Sunrise",             0      },
	{ "SUNSET",         93, 2, GET_time,          'U',
	  "Sunset",              0      },
};

int map_items = sizeof(t) / sizeof(map);

static void *_malloc(size_t size)
{
	void	*p;

	p = malloc(size);
	if (!p) {
		/* FIXME: ... */
		fprintf(stderr,"dwt: malloc failed\n");
		exit(-1);
	}
	return(p);
}


static ushort get_double_byte(char *buf, ushort pos)
{
	 return(((unsigned char)*(buf+pos+1) << 8) | (unsigned char)*(buf+pos));
}

static ushort get_units(char *u)
{
	char	*p,*s;
	ushort	i = 0;

	s = strdup(U_STR);
	p = strtok(s,"|");
	while (p && (strcmp(p,u)!=0)) {
		p = strtok(NULL, "|");
		i++;
	}
	free(s);
	return(p ? i+1: 0);
}


static void print_str(char *frmt, struct dmap *t, char *bp, char *u)
{
	ushort	l, cu;
	float	*v1;
	char	*v2;
	ushort	*v3;

	if (!(u && (cu = get_units(u)))) 
		cu = t->conv;

	l = strlen(frmt);
	switch(t->type) {
		case 'F':
			frmt[l]='f';
			v1 = t->fn(bp, t->ofs, t->len,cu);
			printf(frmt, *v1); 
			free(v1);
			break;

		case 'S':
			frmt[l]='s';
			v2 = t->fn(bp, t->ofs, t->len,cu);
			printf(frmt, v2); 
			free(v2);
			break;

		case 'U':
			frmt[l]='u';
			v3 = t->fn(bp, t->ofs, t->len,cu);
			printf(frmt, *(ushort *)v3); 
			free(v3);
 			break;
	}
	fflush(stdout);
}


static void translate_str(char *src, char *dst)
{
	char	*p1,*p2,*endbuf;

        strcpy(dst,src);
	endbuf = dst+strlen(dst);

        for (p1=src, p2=dst; *p1 && p2<endbuf; ++p1)
        {
                if (*p1 != '\\')
                        *p2++ = *p1;
                else
                {
                        char    c = 0;

                        switch (*++p1)
                        {
                            case 0 :
                                --p1;
                                break;
                            case 'a' :
                                c = 7;
                                break;
                            case 'b' :
                                c = 8;
                                break;
                            case 'f' :
                                c = 12;
                                break;
                            case 'n' :
                                c = 10;
                                break;
                            case 'r' :
                                c = 13;
                                break;
                            case 't' :
                                c = 9;
                                break;
                            case 'v' :
                                c = 11;
                                break;
                            default :
                                if (isdigit(*p1))
                                {
                                        char    oct[4];
                                        ushort     i;

                                        for (i=0; i<3; ++i)
                                                if (isdigit(*p1))
                                                        oct[i] = *p1++;
                                        oct[i] = 0;
                                        c = strtol(oct, 0, 8);
                                        --p1;
                                }
                                else
                                        c = *p1;
                                break;
                        }
                        *p2++ = c;
                }
        }
        *p2 = 0;
}



void *GET_temp(char *buf, ushort ofs, ushort len, ushort conv)
{
	float	*x;

	x = _malloc(sizeof(float));
	*x = (float) get_double_byte(buf, ofs)/10;

	switch(conv) {
		
		case U_CEL :
			*x = (((*x - 32) *5) /9);
			break;

		case U_FAH :
		default:
			break;
	}
	return (x); 
}

void *GET_barometer(char *buf, ushort ofs, ushort len, ushort conv)
{
	float	*x;

	x = _malloc(sizeof(float));
	*x = (float)get_double_byte(buf, ofs);

	switch(conv) {
		case U_HPA :
			*x =   *x * 0.03378;
			break;

		case U_IHG:
		default:
			*x = *x / 1000;
			break;
	}	
	return(x); 
}

void *GET_humidity(char *buf, ushort ofs, ushort len, ushort conv)
{
	ushort	*x;

	x = _malloc(sizeof(ushort));

	*x = *(buf+ofs);

	return (x); 
}

void *GET_rain(char *buf, ushort ofs, ushort len, ushort conv)
{
	float	*f;

	f = _malloc(sizeof(float));
	*f = (float)get_double_byte(buf, ofs);

	switch(conv) {
		case U_MIL :
			*f = *f * 0.254;
			break;

		case U_INC :
			*f = *f * 0.01;
			break;

		case U_CLK :
		default:
			break;
	}
	return(f); 
}

void *GET_windspeed(char *buf, ushort ofs, ushort len, ushort conv)
{
	ushort *x;

	x = _malloc(sizeof(ushort));
	*x = *(buf+ofs);
	switch(conv) {
		case U_KPH:
			*x =  (ushort) ((double) *x * 1.609344  + 0.5);
			break;

		case U_MPS:
			*x =  (ushort) ((double) *x*1.609344*1000.0/3600.0);
			break;

		case U_MPH:
		default:
			break;
	}
	return (x); 
}

void *GET_winddirection(char *buf, ushort ofs, ushort len, ushort conv)
{
	char *w[] = {	"N", "NNE", "NE", "ENE", 
			"E", "ESE", "SE", "SSE",
			"S", "SSW", "SW", "WSW",
			"W", "WNW", "NW", "NNW"  };

        float	r;
	ushort	x;
	char	*p;

        r = (float)get_double_byte(buf, ofs) + 11.25;
	if (r > 360)
		 r = r - 360;
	x = r / 22.5;
	p = strdup(w[x]);

	return(p); 
}

void *GET_winddegree(char *buf, ushort ofs, ushort len, ushort conv)
{

        ushort	*r;

        r = _malloc(sizeof(ushort));
        *r = (ushort)get_double_byte(buf, ofs);

	return (r); 
}

int dwt_make_hash_table(void)
{
        int k;
        ENTRY e, *ep;
	int i;

	k = sizeof(t) / sizeof(map);
        hcreate(k);
        for (i = 0; i < k; i++) {
                e.key = t[i].tag;
                e.data = (char *) i;
                ep = hsearch(e, ENTER);
        }

	return 0;
}

int dwt_destroy_hash_table(void)
{
	hdestroy();
	
	return 0;
}


void *GET_FIELD(char *fld, char *bp, ushort conv)
{
        int	k;
        ENTRY	e, *ep = NULL;
        struct dmap *tp = NULL;

	e.key = fld;
	ep = hsearch(e, FIND);

	if (ep) {
		k = (int)ep->data;
		tp = &t[k];
	}

	return( ep ? tp->fn(bp, tp->ofs, tp->len, conv ? conv : tp->conv) : NULL );
}


void *GET_dewpoint(char *buf, ushort ofs, ushort len, ushort conv)
{
	float	*temp, ews, num1, num2, *r;
	ushort	*hum;


	temp = _malloc(sizeof(float));
	hum = _malloc(sizeof(ushort));
	r = _malloc(sizeof(float));	

	temp = (float *)GET_FIELD("OUT_TEMP", buf, 0);
	hum = (ushort *)GET_FIELD("OUT_HUMIDITY", buf, 0);

	ews = *hum * 0.01 * exp((17.502 * *temp)/(240.9 + *temp));
	num1 = 240.9 * (log(ews));
	num2 = 17.5 - (log(ews));

	*r = num1 / num2;
	free(temp);
	free(hum);


	switch(conv) {

                case U_CEL :
                        break;

                case U_FAH :
                default:
			*r = ((*r * 9) / 5 )+ 32;
                        break;
        }

	return(r);

}

void *GET_windchill(char *buf, ushort ofs, ushort len, ushort conv)
{
        float   *temp, *r;
        ushort  *hum, *windspeed;

        temp = _malloc(sizeof(float));
        hum = _malloc(sizeof(ushort));
        r = _malloc(sizeof(float));
        windspeed = _malloc(sizeof(ushort));

        temp = (float *)GET_FIELD("OUT_TEMP", buf, U_FAH);
        hum = (ushort *)GET_FIELD("OUT_HUMIDITY", buf, 0);
        windspeed = (ushort *)GET_FIELD("WIND_SPEED", buf, U_MPH);

	if (*windspeed < 3)
		*r =  *temp;
	else 
	        *r = 35.74 + (0.6215 * *temp) - (35.75 * expf(0.16*log(*windspeed)))  + (0.4275 * *temp*(expf(0.16*log(*windspeed))));

        switch(conv) {

                case U_CEL :
			*r = (((*r - 32) *5) /9);
                        break;

                case U_FAH :
                default:
                        break;
        }

	free(temp);
	free(hum);
	free(windspeed);

        return(r);

}

void *GET_heatindex(char *buf, ushort ofs, ushort len, ushort conv)
{
        float   *temp, *r, T;
        ushort  *hum, H;

        temp = _malloc(sizeof(float));
        hum = _malloc(sizeof(ushort));
        r = _malloc(sizeof(float));

        temp = (float *)GET_FIELD("OUT_TEMP", buf, U_FAH);
        hum = (ushort *)GET_FIELD("OUT_HUMIDITY", buf, 0);

	T = *temp;
	H = *hum;

	if (T < 80)
		*r = T;
	else
		*r =	-42.379 + (2.04901523 * T) + (10.14333127 * H) - (0.22475541 * T * H) - (0.00683783 * pow(T,2)) - 
			(0.05481717 * pow(H,2)) + (0.00122874 * pow(T,2) * H) + (0.00085282 * T * pow(H,2)) - (0.000001998 * pow(T*H,2));

        switch(conv) {

                case U_CEL :
			*r = (((*r - 32) *5) /9);
                        break;

                case U_FAH :
                default:
                        break;
        }

	free(temp);
	free(hum);

        return(r);

}



void *GET_time(char *buf, ushort ofs, ushort len, ushort conv)
{
	ushort	*x;

        x = _malloc(sizeof(ushort));
        *x = (ushort)get_double_byte(buf, ofs);
		
	return(x);
}


static void proc_frmt_str(char *str, char *bp)
{
        ushort	i, j, n, mode, prn;
	int	k;
        ENTRY	e, *ep;
        char	tbuf[30], *st, *et, *prbuf, *p, *u;
        struct dmap	*tpa;


	i = j = mode = prn = 0;
	st = et = NULL;
	n = strlen(str);
	prbuf = calloc(1,n--);
	while (i <= n) {
		if (!mode && *(str+i) == '%') 
			mode=1;

		if (*(str+i) == '{' && mode) {
			st = str+i;  et = NULL; j = 1;
			while ( i+j <= n) {
				if (*(str+i+j) != '}') {
					j++;
				}
				else {
					et = str+i+j; mode=0;
					break;
				}
			}
			if (et != NULL && et - st > 1) {
				strncpy(tbuf, st+1, et-st-1);
				tbuf[et-st-1]='\0';
				i+=j;
			}
		}
		if (i <= n && !et) {
			prbuf[prn++]=str[i];
		}
		if (i == n && !et) {
			printf(prbuf);
			bzero(prbuf,n+1);
			prn=0;
		}
		if (et) {
			p = strtok(tbuf, "/");
			u = strtok(NULL, "/");
			
			e.key = p;
			ep = hsearch(e, FIND);
			if (ep) {
				k = (int)ep->data;
				tpa = &t[k];
				print_str(prbuf, tpa, bp, u);
			}
			else {
				prbuf[prn]='s';
				printf(prbuf,"-");
			}
			et = NULL;
			bzero(prbuf,n+1);
			prn=0;
		}
		i++;
	}
}

void CMD_list(int fd)
{
        ushort	n;
	int	i;
        char	prbuf[10], *buf;
        ENTRY	e, *ep;

	buf = GET_loop(fd);
	/*
	for(i=0;i<LOOP_LEN;i++)
		printf("%02x ", (unsigned char) buf[i]);
	*/
	if (buf) {
	        n = sizeof(t) / sizeof(map);

		hcreate(n);
	        for (i = 0; i < n; i++) {
			e.key = t[i].tag;
			e.data = (char *)i;
			ep = hsearch(e, ENTER);
	        }

		for (i=0; i<n; i++) {
			printf("\n%20s :  ", t[i].desc);
			bzero(prbuf,10);
			sprintf(prbuf, t[i].type == 'F' ? "%%-10.1" : "%%-10");
			print_str(prbuf, &t[i], buf, NULL);
		}
		printf("\n\n");
		free(buf);
		hdestroy();
	}
}

void CMD_frmt(int fd, char *frmt, ushort n, ushort d)
{
	ushort	k;
	int	i;
	char	str[1024], *buf = NULL;
	ENTRY	e, *ep;

        k = sizeof(t) / sizeof(map);
        hcreate(k);
        for (i = 0; i < k; i++) {
                e.key = t[i].tag;
                e.data = (char *)i;
                ep = hsearch(e, ENTER);
        }

	translate_str(frmt, str);
	for (i = 0; i < n; i++) {
		buf = GET_loop(fd);
		if (buf) {
			proc_frmt_str(str, buf);
			free(buf);
			sleep(d);
		}
	}
	free(buf);
	hdestroy();
}

void CMD_tags()
{
	ushort  k, i;

        k = sizeof(t) / sizeof(map);
	printf("\n%20s  :  %s\n\n", "TAG","DESCRIPTION");
        for (i = 0; i < k; i++) {
                printf("%20s :  %s\n", t[i].tag, t[i].desc);
        }
	printf("\n");
}


ushort get_cmd(char *u)
{
	char	*p,*s;
	ushort	i = 0;

	s = strdup(U_CMD);
	p = strtok(s,"|");
	while (p && strcmp(p,u) != 0) {
		p = strtok(NULL, "|");
		i++;
	}
	free(s);
	return(p ? i+1: 0);
}

ushort valid_cmd(char *cmd)
{
        char	*p,*s;
        ushort	i = 0;

        for (i = 0; i < strlen(cmd); i++)
                cmd[i] = toupper(cmd[i]);

        i = 0;

        s = strdup(U_CMD);
        p = strtok(s, "|");
	do {
		if (strcmp(p, cmd) == 0)
			i = 1;
                p = strtok(NULL, "|");
        } while (p && !i);
        free(s);
        return(i);
}


char *dwt_type(int type)
{
	char *str = (char *) malloc(20);

	/* Davis Weather Station Type: */
	switch(type) {
	case 0:
		sprintf(str, "Wizard III");
		break;
	case 1:
		sprintf(str, "Wizard II");
		break;
	case 2:
		sprintf(str, "Monitor");
		break;
	case 3:
		sprintf(str, "Perception");
		break;
	case 4:
		sprintf(str, "GroWeather");
		break;
	case 5:
		sprintf(str, "Energy Enviromonitor");
		break;
	case 6:
		sprintf(str, "Health Enviromonitor");
		break;
	case 16:
		sprintf(str, "Vantage Pro");
		break;
	default:
		sprintf(str, "Unknown");
		break;
	}

	return str;
}
