/** 
 *   Thermometer 1529 Chub E-4
 *
 *   (c) Copyright 2007 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 *   Based on TC6003P, originally written for Site testing for SKA
 *   candidate.
 *
 *   2007/04/18: Created. (Fede)
 **/

#define _GNU_SOURCE

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <termios.h>
#include <unistd.h>
#include <sys/file.h>
#include <string.h>
#include <stdarg.h>

#include "my_stdio.h"
#include "ther_1529.h"

#define THER_1529_CH 1

static ssize_t soft_read_time_out(struct timeval tv, int fd, char *buf, 
				  size_t count)
{
	fd_set rfds;
	int fd_max = fd + 1;
	int retval, l = 0;
	struct timeval tvl;

	do {
	  memcpy(&tvl, &tv, sizeof(struct timeval));
		FD_ZERO(&rfds);
		FD_SET(fd, &rfds);
		retval = select(fd_max, &rfds, NULL, NULL, &tvl);
		if (retval) {
			if( FD_ISSET(fd, &rfds) ) {
			    retval = read(fd, buf+l, count - l);
				l += retval;
			}
		}else{
			return -1;
		}
	} while ( l < count && *(buf+l-1) != 0x0a) ;

#ifdef DEBUG_MODE
        if (verbose_flag & PINFO) {
                int __i;
                fprintf(stderr, __FILE__ ":read <- ");
                for(__i=0;__i<l;__i++)
                        if( *(((u_char *)buf)+__i) > 0x20) {
                                fprintf(stderr, "%c",
                                      *(((u_char *)buf)+__i) );
                        }else{
                                fprintf(stderr, "[0x%02x]",
                                      *(((u_char *)buf)+__i) );
                        }
                fprintf(stderr, "\n");
        }
#endif
	
	return l;
}


static int my_write(int dev, const char *fmt, ...)
{
        va_list ap;
        int l = -1;
        int n, size = 80;
        char *buff = NULL;

        if ((buff = malloc (size)) == NULL) return l;
        while (1) {
                /* Try to print in the allocated space. */
                va_start(ap, fmt);
                n = vsnprintf (buff, size, fmt, ap);
                va_end(ap);
                if (n > -1 && n < size) {
                        break;
                }
                else {
                        /* Else try again with more space. */
                        if (n > -1)    /* glibc 2.1 */
                                size = n+1; /* precisely what is needed */
                        else           /* glibc 2.0 */
                                size *= 2;  /* twice the old size */
                        if ((buff = realloc (buff, size)) == NULL)
                                return l;
                }
        }

        if (buff != NULL) {
                PVERB(PINFO,"> ther1529(%d): %s", dev, buff);
                l = write(dev, buff, n);
                free(buff);
        }

        return l;
}



int ther1529_open(char *name)
{
	int fd;
	struct termios tio;

	/* Open Serial Port : 8 bits data, 1 bit stop, Even parity, 19200
           baud */
	if ((fd = open(name, O_RDWR | O_NOCTTY)) < 0 ){
		SYSLOG(LOG_ERR, "Cannot open serial port (%s).", name);
		return -1;
	}

	memset(&tio, 0, sizeof(tio));
	/* Ignore BREAK condition on input */
	tio.c_iflag |= IGNBRK;
	tio.c_cflag |= (B19200|CRTSCTS|CLOCAL|CREAD|CS8);
	/* Set MIN to one and TIME to zero */
	tio.c_cc[VMIN]  = 1;
	tio.c_cc[VTIME] = 0;

	/* now clean the modem line and activate the settings for modem */
	tcflush(fd, TCOFLUSH|TCIFLUSH);
	tcsetattr(fd, TCSANOW, &tio);

	my_write(fd, "UNIT:TEMP C\n");
	//my_write(fd, "SYST:KLOC 1\n");

	return fd;
}


int ther1529_close(int fd)
{
	//my_write(fd, "SYST:KLOC 0\n");

	return close(fd);
}


int ther1529_get(int fd, double *temperature)
{
	char recv[16];
	struct timeval tv;

	my_write(fd, "fetch?%d\n", THER_1529_CH);
	tv.tv_sec = 0;
	tv.tv_usec = 100000;
	if (soft_read_time_out(tv, fd, recv, 10) < 0) {
		SYSLOG(LOG_ERR, "Thermometer (%d) time out.", fd);
		return THERMOMETER_FAIL;
	}

	*temperature = atof(recv);

	return 0;
}

